@@ -33,7 +33,7 @@ except NameError:
 # This is used to gate several fields/features that only exist
 # for the protobuf_unittest version of the message.
 def IsProto2(message):
-  return message.DESCRIPTOR.full_name == 'protobuf_unittest.TestAllTypes'
+  return message.DESCRIPTOR.full_name == 'proto2_unittest.TestAllTypes'
 
 
 def SetAllNonLazyFields(message):
@@ -77,6 +77,7 @@ def SetAllNonLazyFields(message):
 
   message.optional_string_piece = u'124'
   message.optional_cord = u'125'
+  message.optional_bytes_cord = b'optional bytes cord'
 
   #
   # Repeated fields.
@@ -247,6 +248,7 @@ def SetAllExtensions(message):
 
   extensions[pb2.optional_string_piece_extension] = u'124'
   extensions[pb2.optional_cord_extension] = u'125'
+  extensions[pb2.optional_bytes_cord_extension] = b'optional bytes cord'
 
   #
   # Repeated fields.
@@ -423,6 +425,7 @@ def ExpectAllFieldsSet(test_case, message):
 
   test_case.assertTrue(message.HasField('optional_string_piece'))
   test_case.assertTrue(message.HasField('optional_cord'))
+  test_case.assertTrue(message.HasField('optional_bytes_cord'))
 
   test_case.assertEqual(101, message.optional_int32)
   test_case.assertEqual(102, message.optional_int64)
