@@ -13,16 +13,23 @@
 #include <cstdint>
 #include <limits>
 #include <string>
-#include <typeinfo>
 #include <vector>
 
 #include "absl/base/attributes.h"
+#include "absl/base/dynamic_annotations.h"
+#include "absl/base/optimization.h"
+#include "absl/base/prefetch.h"
 #include "absl/container/internal/layout.h"
+#include "absl/log/absl_check.h"
+#include "absl/log/absl_log.h"
 #include "absl/synchronization/mutex.h"
+#include "absl/types/span.h"
 #include "google/protobuf/arena_allocation_policy.h"
+#include "google/protobuf/arena_cleanup.h"
 #include "google/protobuf/arenaz_sampler.h"
 #include "google/protobuf/port.h"
 #include "google/protobuf/serial_arena.h"
+#include "google/protobuf/string_block.h"
 #include "google/protobuf/thread_safe_arena.h"
 
 
@@ -55,54 +62,146 @@ ArenaBlock* SentryArenaBlock() {
 }
 #endif
 
-}  // namespace
-
-static SizedPtr AllocateMemory(const AllocationPolicy* policy_ptr,
-                               size_t last_size, size_t min_bytes) {
-  AllocationPolicy policy;  // default policy
-  if (policy_ptr) policy = *policy_ptr;
-  size_t size;
-  if (last_size != 0) {
-    // Double the current block size, up to a limit.
-    auto max_size = policy.max_block_size;
-    size = std::min(2 * last_size, max_size);
-  } else {
-    size = policy.start_block_size;
-  }
-  // Verify that min_bytes + kBlockHeaderSize won't overflow.
-  ABSL_CHECK_LE(min_bytes, std::numeric_limits<size_t>::max() -
-                               SerialArena::kBlockHeaderSize);
-  size = std::max(size, SerialArena::kBlockHeaderSize + min_bytes);
+inline size_t AllocationSize(size_t last_size, size_t start_size,
+                             size_t max_size) {
+  if (last_size == 0) return start_size;
+  // Double the current block size, up to a limit.
+  return std::min(2 * last_size, max_size);
+}
 
+SizedPtr AllocateMemory(const AllocationPolicy& policy, size_t size) {
   if (policy.block_alloc == nullptr) {
     return AllocateAtLeast(size);
   }
   return {policy.block_alloc(size), size};
 }
 
+SizedPtr AllocateBlock(const AllocationPolicy* policy_ptr, size_t last_size,
+                       size_t min_bytes) {
+  AllocationPolicy policy;  // default policy
+  if (policy_ptr) policy = *policy_ptr;
+  size_t size =
+      AllocationSize(last_size, policy.start_block_size, policy.max_block_size);
+  // Verify that min_bytes + kBlockHeaderSize won't overflow.
+  ABSL_CHECK_LE(min_bytes, std::numeric_limits<size_t>::max() -
+                               SerialArena::kBlockHeaderSize);
+  size = std::max(size, SerialArena::kBlockHeaderSize + min_bytes);
+
+  return AllocateMemory(policy, size);
+}
+
+SizedPtr AllocateCleanupChunk(const AllocationPolicy* policy_ptr,
+                              size_t last_size) {
+  constexpr size_t kStartSize = 64;
+  constexpr size_t kMaxSize = 4 << 10;
+  static_assert(kStartSize % sizeof(cleanup::CleanupNode) == 0, "");
+
+  const size_t size = AllocationSize(last_size, kStartSize, kMaxSize);
+  if (policy_ptr == nullptr) return AllocateAtLeast(size);
+  return AllocateMemory(*policy_ptr, size);
+}
+
 class GetDeallocator {
  public:
-  GetDeallocator(const AllocationPolicy* policy, size_t* space_allocated)
-      : dealloc_(policy ? policy->block_dealloc : nullptr),
-        space_allocated_(space_allocated) {}
+  explicit GetDeallocator(const AllocationPolicy* policy)
+      : dealloc_(policy ? policy->block_dealloc : nullptr) {}
 
   void operator()(SizedPtr mem) const {
-    // This memory was provided by the underlying allocator as unpoisoned,
-    // so return it in an unpoisoned state.
-    PROTOBUF_UNPOISON_MEMORY_REGION(mem.p, mem.n);
     if (dealloc_) {
       dealloc_(mem.p, mem.n);
     } else {
       internal::SizedDelete(mem.p, mem.n);
     }
-    *space_allocated_ += mem.n;
   }
 
  private:
   void (*dealloc_)(void*, size_t);
-  size_t* space_allocated_;
 };
 
+}  // namespace
+
+namespace cleanup {
+struct ChunkList::Chunk {
+  CleanupNode* First() { return reinterpret_cast<CleanupNode*>(this + 1); }
+  CleanupNode* Last() { return First() + Capacity() - 1; }
+  static size_t Capacity(size_t size) {
+    return (size - sizeof(Chunk)) / sizeof(CleanupNode);
+  }
+  size_t Capacity() const { return Capacity(size); }
+
+  Chunk* next;
+  size_t size;
+  // Cleanup nodes follow.
+};
+
+void ChunkList::AddFallback(void* elem, void (*destructor)(void*),
+                            SerialArena& arena) {
+  ABSL_DCHECK_EQ(next_, limit_);
+  SizedPtr mem = AllocateCleanupChunk(arena.parent_.AllocPolicy(),
+                                      head_ == nullptr ? 0 : head_->size);
+  arena.AddSpaceAllocated(mem.n);
+  head_ = new (mem.p) Chunk{head_, mem.n};
+  next_ = head_->First();
+  prefetch_ptr_ = reinterpret_cast<char*>(next_);
+  limit_ = next_ + Chunk::Capacity(mem.n);
+  AddFromExisting(elem, destructor);
+}
+
+void ChunkList::Cleanup(const SerialArena& arena) {
+  Chunk* c = head_;
+  if (c == nullptr) return;
+  GetDeallocator deallocator(arena.parent_.AllocPolicy());
+
+  // Iterate backwards in order to destroy in the right order.
+  CleanupNode* it = next_ - 1;
+  while (true) {
+    CleanupNode* first = c->First();
+    // A prefetch distance of 8 here was chosen arbitrarily.
+    constexpr int kPrefetchDistance = 8;
+    CleanupNode* prefetch = it;
+    // Prefetch the first kPrefetchDistance nodes.
+    for (int i = 0; prefetch >= first && i < kPrefetchDistance;
+         --prefetch, ++i) {
+      prefetch->Prefetch();
+    }
+    // For the middle nodes, run destructor and prefetch the node
+    // kPrefetchDistance after the current one.
+    for (; prefetch >= first; --it, --prefetch) {
+      it->Destroy();
+      prefetch->Prefetch();
+    }
+    // Note: we could consider prefetching `next` chunk earlier.
+    absl::PrefetchToLocalCacheNta(c->next);
+    // Destroy the rest without prefetching.
+    for (; it >= first; --it) {
+      it->Destroy();
+    }
+    Chunk* next = c->next;
+    deallocator({c, c->size});
+    if (next == nullptr) return;
+    c = next;
+    it = c->Last();
+  };
+}
+
+std::vector<void*> ChunkList::PeekForTesting() {
+  std::vector<void*> ret;
+  Chunk* c = head_;
+  if (c == nullptr) return ret;
+  // Iterate backwards to match destruction order.
+  CleanupNode* it = next_ - 1;
+  while (true) {
+    CleanupNode* first = c->First();
+    for (; it >= first; --it) {
+      ret.push_back(it->elem);
+    }
+    c = c->next;
+    if (c == nullptr) return ret;
+    it = c->Last();
+  };
+}
+}  // namespace cleanup
+
 // It is guaranteed that this is constructed in `b`. IOW, this is not the first
 // arena and `b` cannot be sentry.
 SerialArena::SerialArena(ArenaBlock* b, ThreadSafeArena& parent)
@@ -110,7 +209,6 @@ SerialArena::SerialArena(ArenaBlock* b, ThreadSafeArena& parent)
       limit_{b->Limit()},
       prefetch_ptr_(
           b->Pointer(kBlockHeaderSize + ThreadSafeArena::kSerialArenaSize)),
-      prefetch_limit_(b->Limit()),
       head_{b},
       space_allocated_{b->size},
       parent_{parent} {
@@ -131,22 +229,7 @@ SerialArena::SerialArena(FirstSerialArena, ArenaBlock* b,
 }
 
 std::vector<void*> SerialArena::PeekCleanupListForTesting() {
-  std::vector<void*> res;
-
-  ArenaBlock* b = head();
-  if (b->IsSentry()) return res;
-
-  const auto peek_list = [&](const char* pos, const char* end) {
-    while (pos != end) {
-      pos += cleanup::PeekNode(pos, res);
-    }
-  };
-
-  peek_list(limit_, b->Limit());
-  for (b = b->next; b; b = b->next) {
-    peek_list(reinterpret_cast<char*>(b->cleanup_nodes), b->Limit());
-  }
-  return res;
+  return cleanup_list_.PeekForTesting();
 }
 
 std::vector<void*> ThreadSafeArena::PeekCleanupListForTesting() {
@@ -175,6 +258,8 @@ SerialArena* SerialArena::New(SizedPtr mem, ThreadSafeArena& parent) {
 
 template <typename Deallocator>
 SizedPtr SerialArena::Free(Deallocator deallocator) {
+  FreeStringBlocks();
+
   ArenaBlock* b = head();
   SizedPtr mem = {b, b->size};
   while (b->next) {
@@ -188,7 +273,7 @@ SizedPtr SerialArena::Free(Deallocator deallocator) {
 PROTOBUF_NOINLINE
 void* SerialArena::AllocateAlignedFallback(size_t n) {
   AllocateNewBlock(n);
-  void* ret;
+  void* ret = nullptr;
   bool res = MaybeAllocateAligned(n, &ret);
   ABSL_DCHECK(res);
   return ret;
@@ -222,26 +307,16 @@ void* SerialArena::AllocateFromStringBlockFallback() {
 PROTOBUF_NOINLINE
 void* SerialArena::AllocateAlignedWithCleanupFallback(
     size_t n, size_t align, void (*destructor)(void*)) {
-  size_t required = AlignUpTo(n, align) + cleanup::Size(destructor);
+  size_t required = AlignUpTo(n, align);
   AllocateNewBlock(required);
   return AllocateAlignedWithCleanup(n, align, destructor);
 }
 
-PROTOBUF_NOINLINE
-void SerialArena::AddCleanupFallback(void* elem, void (*destructor)(void*)) {
-  size_t required = cleanup::Size(destructor);
-  AllocateNewBlock(required);
-  AddCleanupFromExisting(elem, destructor);
-}
-
 void SerialArena::AllocateNewBlock(size_t n) {
   size_t used = 0;
   size_t wasted = 0;
   ArenaBlock* old_head = head();
   if (!old_head->IsSentry()) {
-    // Sync limit to block
-    old_head->cleanup_nodes = limit_;
-
     // Record how much used in this block.
     used = static_cast<size_t>(ptr() - old_head->Pointer(kBlockHeaderSize));
     wasted = old_head->size - used - kBlockHeaderSize;
@@ -253,10 +328,7 @@ void SerialArena::AllocateNewBlock(size_t n) {
   // but with a CPU regression. The regression might have been an artifact of
   // the microbenchmark.
 
-  auto mem = AllocateMemory(parent_.AllocPolicy(), old_head->size, n);
-  // We don't want to emit an expensive RMW instruction that requires
-  // exclusive access to a cacheline. Hence we write it in terms of a
-  // regular add.
+  auto mem = AllocateBlock(parent_.AllocPolicy(), old_head->size, n);
   AddSpaceAllocated(mem.n);
   ThreadSafeArenaStats::RecordAllocateStats(parent_.arena_stats_.MutableStats(),
                                             /*used=*/used,
@@ -266,7 +338,7 @@ void SerialArena::AllocateNewBlock(size_t n) {
   // Previous writes must take effect before writing new head.
   head_.store(new_head, std::memory_order_release);
 
-  PROTOBUF_POISON_MEMORY_REGION(ptr(), limit_ - ptr());
+  internal::PoisonMemoryRegion(ptr(), limit_ - ptr());
 }
 
 uint64_t SerialArena::SpaceUsed() const {
@@ -299,6 +371,7 @@ size_t SerialArena::FreeStringBlocks(StringBlock* string_block,
                                      size_t unused_bytes) {
   ABSL_DCHECK(string_block != nullptr);
   StringBlock* next = string_block->next();
+  absl::PrefetchToLocalCacheNta(next);
   std::string* end = string_block->end();
   for (std::string* s = string_block->AtOffset(unused_bytes); s != end; ++s) {
     s->~basic_string();
@@ -307,6 +380,7 @@ size_t SerialArena::FreeStringBlocks(StringBlock* string_block,
 
   while ((string_block = next) != nullptr) {
     next = string_block->next();
+    absl::PrefetchToLocalCacheNta(next);
     for (std::string& s : *string_block) {
       s.~basic_string();
     }
@@ -315,22 +389,6 @@ size_t SerialArena::FreeStringBlocks(StringBlock* string_block,
   return deallocated;
 }
 
-void SerialArena::CleanupList() {
-  ArenaBlock* b = head();
-  if (b->IsSentry()) return;
-
-  b->cleanup_nodes = limit_;
-  do {
-    char* limit = b->Limit();
-    char* it = reinterpret_cast<char*>(b->cleanup_nodes);
-    ABSL_DCHECK(!b->IsSentry() || it == limit);
-    while (it < limit) {
-      it += cleanup::DestroyNode(it);
-    }
-    b = b->next;
-  } while (b);
-}
-
 // Stores arrays of void* and SerialArena* instead of linked list of
 // SerialArena* to speed up traversing all SerialArena. The cost of walk is non
 // trivial when there are many nodes. Separately storing "ids" minimizes cache
@@ -362,16 +420,21 @@ struct SerialArenaChunkHeader {
 class ThreadSafeArena::SerialArenaChunk {
  public:
   SerialArenaChunk(uint32_t capacity, void* me, SerialArena* serial) {
-    new (&header()) SerialArenaChunkHeader{capacity, 1};
+    // We use `layout`/`ids`/`arenas` local variables to avoid recomputing
+    // offsets if we were to call id(i)/arena(i) repeatedly.
+    const layout_type layout = Layout(capacity);
+    new (layout.Pointer<kHeader>(ptr())) SerialArenaChunkHeader{capacity, 1};
 
-    new (&id(0)) std::atomic<void*>{me};
+    std::atomic<void*>* ids = layout.Pointer<kIds>(ptr());
+    new (&ids[0]) std::atomic<void*>{me};
     for (uint32_t i = 1; i < capacity; ++i) {
-      new (&id(i)) std::atomic<void*>{nullptr};
+      new (&ids[i]) std::atomic<void*>{nullptr};
     }
 
-    new (&arena(0)) std::atomic<SerialArena*>{serial};
+    std::atomic<SerialArena*>* arenas = layout.Pointer<kArenas>(ptr());
+    new (&arenas[0]) std::atomic<SerialArena*>{serial};
     for (uint32_t i = 1; i < capacity; ++i) {
-      new (&arena(i)) std::atomic<void*>{nullptr};
+      new (&arenas[i]) std::atomic<SerialArena*>{nullptr};
     }
   }
 
@@ -390,30 +453,30 @@ class ThreadSafeArena::SerialArenaChunk {
 
   // ids: returns up to size().
   absl::Span<const std::atomic<void*>> ids() const {
-    return Layout(capacity()).Slice<kIds>(ptr()).first(safe_size());
+    return Layout().Slice<kIds>(ptr()).first(safe_size());
   }
   absl::Span<std::atomic<void*>> ids() {
-    return Layout(capacity()).Slice<kIds>(ptr()).first(safe_size());
+    return Layout().Slice<kIds>(ptr()).first(safe_size());
   }
   std::atomic<void*>& id(uint32_t i) {
     ABSL_DCHECK_LT(i, capacity());
-    return Layout(capacity()).Pointer<kIds>(ptr())[i];
+    return Layout().Pointer<kIds>(ptr())[i];
   }
 
   // arenas: returns up to size().
   absl::Span<const std::atomic<SerialArena*>> arenas() const {
-    return Layout(capacity()).Slice<kArenas>(ptr()).first(safe_size());
+    return Layout().Slice<kArenas>(ptr()).first(safe_size());
   }
   absl::Span<std::atomic<SerialArena*>> arenas() {
-    return Layout(capacity()).Slice<kArenas>(ptr()).first(safe_size());
+    return Layout().Slice<kArenas>(ptr()).first(safe_size());
   }
   const std::atomic<SerialArena*>& arena(uint32_t i) const {
     ABSL_DCHECK_LT(i, capacity());
-    return Layout(capacity()).Pointer<kArenas>(ptr())[i];
+    return Layout().Pointer<kArenas>(ptr())[i];
   }
   std::atomic<SerialArena*>& arena(uint32_t i) {
     ABSL_DCHECK_LT(i, capacity());
-    return Layout(capacity()).Pointer<kArenas>(ptr())[i];
+    return Layout().Pointer<kArenas>(ptr())[i];
   }
 
   // Tries to insert {id, serial} to head chunk. Returns false if the head is
@@ -446,7 +509,8 @@ class ThreadSafeArena::SerialArenaChunk {
   constexpr static int kArenas = 2;
 
   using layout_type = absl::container_internal::Layout<
-      SerialArenaChunkHeader, std::atomic<void*>, std::atomic<SerialArena*>>;
+      SerialArenaChunkHeader, std::atomic<void*>,
+      std::atomic<SerialArena*>>::WithStaticSizes</*header*/ 1>;
 
   const char* ptr() const { return reinterpret_cast<const char*>(this); }
   char* ptr() { return reinterpret_cast<char*>(this); }
@@ -468,11 +532,9 @@ class ThreadSafeArena::SerialArenaChunk {
   }
 
   constexpr static layout_type Layout(size_t n) {
-    return layout_type(
-        /*header*/ 1,
-        /*ids*/ n,
-        /*arenas*/ n);
+    return layout_type(/*ids*/ n, /*arenas*/ n);
   }
+  layout_type Layout() const { return Layout(capacity()); }
 };
 
 constexpr SerialArenaChunkHeader kSentryArenaChunk = {0, 0};
@@ -524,6 +586,16 @@ ArenaBlock* ThreadSafeArena::FirstBlock(void* buf, size_t size) {
     return SentryArenaBlock();
   }
   // Record user-owned block.
+
+  if constexpr (internal::PerformDebugChecks()) {
+    // Touch block to verify it is addressable.
+    if (size > 0) {
+      static_cast<char*>(buf)[0] = 0;
+      static_cast<char*>(buf)[size - 1] = 0;
+    }
+  }
+
+  ABSL_ANNOTATE_MEMORY_IS_UNINITIALIZED(buf, size);
   alloc_policy_.set_is_user_owned_initial_block(true);
   return new (buf) ArenaBlock{nullptr, size};
 }
@@ -536,10 +608,18 @@ ArenaBlock* ThreadSafeArena::FirstBlock(void* buf, size_t size,
 
   SizedPtr mem;
   if (buf == nullptr || size < kBlockHeaderSize + kAllocPolicySize) {
-    mem = AllocateMemory(&policy, 0, kAllocPolicySize);
+    mem = AllocateBlock(&policy, 0, kAllocPolicySize);
   } else {
     mem = {buf, size};
     // Record user-owned block.
+    if constexpr (internal::PerformDebugChecks()) {
+      // Touch block to verify it is addressable.
+      if (size > 0) {
+        static_cast<char*>(buf)[0] = 0;
+        static_cast<char*>(buf)[size - 1] = 0;
+      }
+    }
+    ABSL_ANNOTATE_MEMORY_IS_UNINITIALIZED(buf, size);
     alloc_policy_.set_is_user_owned_initial_block(true);
   }
 
@@ -580,7 +660,7 @@ uint64_t ThreadSafeArena::GetNextLifeCycleId() {
   ThreadCache& tc = thread_cache();
   uint64_t id = tc.next_lifecycle_id;
   constexpr uint64_t kInc = ThreadCache::kPerThreadIds;
-  if (PROTOBUF_PREDICT_FALSE((id & (kInc - 1)) == 0)) {
+  if (ABSL_PREDICT_FALSE((id & (kInc - 1)) == 0)) {
     // On platforms that don't support uint64_t atomics we can certainly not
     // afford to increment by large intervals and expect uniqueness due to
     // wrapping, hence we only add by 1.
@@ -648,6 +728,15 @@ void ThreadSafeArena::AddSerialArena(void* id, SerialArena* serial) {
   head_.store(new_head, std::memory_order_release);
 }
 
+void ThreadSafeArena::UnpoisonAllArenaBlocks() const {
+  VisitSerialArena([](const SerialArena* serial) {
+    for (const auto* b = serial->head(); b != nullptr && !b->IsSentry();
+         b = b->next) {
+      internal::UnpoisonMemoryRegion(b, b->size);
+    }
+  });
+}
+
 void ThreadSafeArena::Init() {
   tag_and_id_ = GetNextLifeCycleId();
   arena_stats_ = Sample();
@@ -670,19 +759,17 @@ ThreadSafeArena::~ThreadSafeArena() {
   // refer to memory in other blocks.
   CleanupList();
 
-  size_t space_allocated = 0;
-  auto mem = Free(&space_allocated);
+  auto mem = Free();
   if (alloc_policy_.is_user_owned_initial_block()) {
     // Unpoison the initial block, now that it's going back to the user.
-    PROTOBUF_UNPOISON_MEMORY_REGION(mem.p, mem.n);
-    space_allocated += mem.n;
+    internal::UnpoisonMemoryRegion(mem.p, mem.n);
   } else if (mem.n > 0) {
-    GetDeallocator(alloc_policy_.get(), &space_allocated)(mem);
+    GetDeallocator(alloc_policy_.get())(mem);
   }
 }
 
-SizedPtr ThreadSafeArena::Free(size_t* space_allocated) {
-  auto deallocator = GetDeallocator(alloc_policy_.get(), space_allocated);
+SizedPtr ThreadSafeArena::Free() {
+  auto deallocator = GetDeallocator(alloc_policy_.get());
 
   WalkSerialArenaChunk([&](SerialArenaChunk* chunk) {
     absl::Span<std::atomic<SerialArena*>> span = chunk->arenas();
@@ -692,8 +779,6 @@ SizedPtr ThreadSafeArena::Free(size_t* space_allocated) {
     for (auto it = span.rbegin(); it != span.rend(); ++it) {
       SerialArena* serial = it->load(std::memory_order_relaxed);
       ABSL_DCHECK_NE(serial, nullptr);
-      // Free string blocks
-      *space_allocated += serial->FreeStringBlocks();
       // Always frees the first block of "serial" as it cannot be user-provided.
       SizedPtr mem = serial->Free(deallocator);
       ABSL_DCHECK_NE(mem.p, nullptr);
@@ -706,20 +791,21 @@ SizedPtr ThreadSafeArena::Free(size_t* space_allocated) {
   });
 
   // The first block of the first arena is special and let the caller handle it.
-  *space_allocated += first_arena_.FreeStringBlocks();
   return first_arena_.Free(deallocator);
 }
 
 uint64_t ThreadSafeArena::Reset() {
+  const size_t space_allocated = SpaceAllocated();
+
   // Have to do this in a first pass, because some of the destructors might
   // refer to memory in other blocks.
   CleanupList();
+  // Reset the first arena's cleanup list.
+  first_arena_.cleanup_list_ = cleanup::ChunkList();
 
   // Discard all blocks except the first one. Whether it is user-provided or
   // allocated, always reuse the first block for the first arena.
-  size_t space_allocated = 0;
-  auto mem = Free(&space_allocated);
-  space_allocated += mem.n;
+  auto mem = Free();
 
   // Reset the first arena with the first block. This avoids redundant
   // free / allocation and re-allocating for AllocationPolicy. Adjust offset if
@@ -729,6 +815,8 @@ uint64_t ThreadSafeArena::Reset() {
     size_t offset = alloc_policy_.get() == nullptr
                         ? kBlockHeaderSize
                         : kBlockHeaderSize + kAllocPolicySize;
+    ABSL_ANNOTATE_MEMORY_IS_UNINITIALIZED(static_cast<char*>(mem.p) + offset,
+                                          mem.n - offset);
     first_arena_.Init(new (mem.p) ArenaBlock{nullptr, mem.n}, offset);
   } else {
     first_arena_.Init(SentryArenaBlock(), 0);
@@ -744,7 +832,7 @@ uint64_t ThreadSafeArena::Reset() {
 void* ThreadSafeArena::AllocateAlignedWithCleanup(size_t n, size_t align,
                                                   void (*destructor)(void*)) {
   SerialArena* arena;
-  if (PROTOBUF_PREDICT_TRUE(GetSerialArenaFast(&arena))) {
+  if (ABSL_PREDICT_TRUE(GetSerialArenaFast(&arena))) {
     return arena->AllocateAlignedWithCleanup(n, align, destructor);
   } else {
     return AllocateAlignedWithCleanupFallback(n, align, destructor);
@@ -757,7 +845,7 @@ void ThreadSafeArena::AddCleanup(void* elem, void (*cleanup)(void*)) {
 
 SerialArena* ThreadSafeArena::GetSerialArena() {
   SerialArena* arena;
-  if (PROTOBUF_PREDICT_FALSE(!GetSerialArenaFast(&arena))) {
+  if (ABSL_PREDICT_FALSE(!GetSerialArenaFast(&arena))) {
     arena = GetSerialArenaFallback(kMaxCleanupNodeSize);
   }
   return arena;
@@ -775,17 +863,19 @@ void* ThreadSafeArena::AllocateFromStringBlock() {
   return GetSerialArena()->AllocateFromStringBlock();
 }
 
-template <typename Functor>
-void ThreadSafeArena::WalkConstSerialArenaChunk(Functor fn) const {
+template <typename Callback>
+void ThreadSafeArena::WalkConstSerialArenaChunk(Callback fn) const {
   const SerialArenaChunk* chunk = head_.load(std::memory_order_acquire);
 
   for (; !chunk->IsSentry(); chunk = chunk->next_chunk()) {
+    // Prefetch the next chunk.
+    absl::PrefetchToLocalCache(chunk->next_chunk());
     fn(chunk);
   }
 }
 
-template <typename Functor>
-void ThreadSafeArena::WalkSerialArenaChunk(Functor fn) {
+template <typename Callback>
+void ThreadSafeArena::WalkSerialArenaChunk(Callback fn) {
   // By omitting an Acquire barrier we help the sanitizer that any user code
   // that doesn't properly synchronize Reset() or the destructor will throw a
   // TSAN warning.
@@ -794,13 +884,19 @@ void ThreadSafeArena::WalkSerialArenaChunk(Functor fn) {
   while (!chunk->IsSentry()) {
     // Cache next chunk in case this chunk is destroyed.
     SerialArenaChunk* next_chunk = chunk->next_chunk();
+    // Prefetch the next chunk.
+    absl::PrefetchToLocalCache(next_chunk);
     fn(chunk);
     chunk = next_chunk;
   }
 }
 
-template <typename Functor>
-void ThreadSafeArena::PerConstSerialArenaInChunk(Functor fn) const {
+template <typename Callback>
+void ThreadSafeArena::VisitSerialArena(Callback fn) const {
+  // In most cases, arenas are single-threaded and "first_arena_" should be
+  // sufficient.
+  fn(&first_arena_);
+
   WalkConstSerialArenaChunk([&fn](const SerialArenaChunk* chunk) {
     for (const auto& each : chunk->arenas()) {
       const SerialArena* serial = each.load(std::memory_order_acquire);
@@ -813,18 +909,17 @@ void ThreadSafeArena::PerConstSerialArenaInChunk(Functor fn) const {
 }
 
 uint64_t ThreadSafeArena::SpaceAllocated() const {
-  uint64_t space_allocated = first_arena_.SpaceAllocated();
-  PerConstSerialArenaInChunk([&space_allocated](const SerialArena* serial) {
+  uint64_t space_allocated = 0;
+  VisitSerialArena([&space_allocated](const SerialArena* serial) {
     space_allocated += serial->SpaceAllocated();
   });
   return space_allocated;
 }
 
 uint64_t ThreadSafeArena::SpaceUsed() const {
-  // First arena is inlined to ThreadSafeArena and the first block's overhead is
-  // smaller than others that contain SerialArena.
-  uint64_t space_used = first_arena_.SpaceUsed();
-  PerConstSerialArenaInChunk([&space_used](const SerialArena* serial) {
+  // `first_arena_` doesn't have kSerialArenaSize overhead, so adjust it here.
+  uint64_t space_used = kSerialArenaSize;
+  VisitSerialArena([&space_used](const SerialArena* serial) {
     // SerialArena on chunks directly allocated from the block and needs to be
     // subtracted from SpaceUsed.
     space_used += serial->SpaceUsed() - kSerialArenaSize;
@@ -843,6 +938,10 @@ template void*
     ThreadSafeArena::AllocateAlignedFallback<AllocationClient::kArray>(size_t);
 
 void ThreadSafeArena::CleanupList() {
+  if constexpr (HasMemoryPoisoning()) {
+    UnpoisonAllArenaBlocks();
+  }
+
   WalkSerialArenaChunk([](SerialArenaChunk* chunk) {
     absl::Span<std::atomic<SerialArena*>> span = chunk->arenas();
     // Walks arenas backward to handle the first serial arena the last.
@@ -884,7 +983,7 @@ SerialArena* ThreadSafeArena::GetSerialArenaFallback(size_t n) {
     // have any blocks yet.  So we'll allocate its first block now. It must be
     // big enough to host SerialArena and the pending request.
     serial = SerialArena::New(
-        AllocateMemory(alloc_policy_.get(), 0, n + kSerialArenaSize), *this);
+        AllocateBlock(alloc_policy_.get(), 0, n + kSerialArenaSize), *this);
 
     AddSerialArena(id, serial);
   }
