@@ -47,7 +47,6 @@ void SetEnumVariables(
 
   (*variables)["type"] =
       name_resolver->GetImmutableClassName(descriptor->enum_type());
-  variables->insert({"kt_type", EscapeKotlinKeywords((*variables)["type"])});
   (*variables)["mutable_type"] =
       name_resolver->GetMutableClassName(descriptor->enum_type());
   (*variables)["default"] =
@@ -62,15 +61,8 @@ void SetEnumVariables(
   // by the proto compiler
   (*variables)["deprecation"] =
       descriptor->options().deprecated() ? "@java.lang.Deprecated " : "";
-  variables->insert(
-      {"kt_deprecation",
-       descriptor->options().deprecated()
-           ? absl::StrCat("@kotlin.Deprecated(message = \"Field ",
-                          (*variables)["name"], " is deprecated\") ")
-           : ""});
   if (HasHasbit(descriptor)) {
     // For singular messages and builders, one bit is used for the hasField bit.
-    (*variables)["get_has_field_bit_message"] = GenerateGetBit(messageBitIndex);
     // Note that these have a trailing ";".
     (*variables)["set_has_field_bit_message"] =
         absl::StrCat(GenerateSetBit(messageBitIndex), ";");
@@ -169,7 +161,7 @@ void ImmutableEnumFieldGenerator::GenerateMembers(io::Printer* printer) const {
     printer->Print(variables_,
                    "@java.lang.Override $deprecation$public boolean "
                    "${$has$capitalized_name$$}$() {\n"
-                   "  return $get_has_field_bit_message$;\n"
+                   "  return $is_field_present_message$;\n"
                    "}\n");
     printer->Annotate("{", "}", descriptor_);
   }
@@ -267,51 +259,6 @@ void ImmutableEnumFieldGenerator::GenerateBuilderMembers(
   printer->Annotate("{", "}", descriptor_, Semantic::kSet);
 }
 
-void ImmutableEnumFieldGenerator::GenerateKotlinDslMembers(
-    io::Printer* printer) const {
-  WriteFieldDocComment(printer, descriptor_, context_->options(),
-                       /* kdoc */ true);
-  printer->Print(variables_,
-                 "$kt_deprecation$public var $kt_name$: $kt_type$\n"
-                 "  @JvmName(\"${$get$kt_capitalized_name$$}$\")\n"
-                 "  get() = $kt_dsl_builder$.${$get$capitalized_name$$}$()\n"
-                 "  @JvmName(\"${$set$kt_capitalized_name$$}$\")\n"
-                 "  set(value) {\n"
-                 "    $kt_dsl_builder$.${$set$capitalized_name$$}$(value)\n"
-                 "  }\n");
-
-  if (SupportUnknownEnumValue(descriptor_)) {
-    printer->Print(
-        variables_,
-        "$kt_deprecation$public var $kt_name$Value: kotlin.Int\n"
-        "  @JvmName(\"${$get$kt_capitalized_name$Value$}$\")\n"
-        "  get() = $kt_dsl_builder$.${$get$capitalized_name$Value$}$()\n"
-        "  @JvmName(\"${$set$kt_capitalized_name$Value$}$\")\n"
-        "  set(value) {\n"
-        "    $kt_dsl_builder$.${$set$capitalized_name$Value$}$(value)\n"
-        "  }\n");
-  }
-
-  WriteFieldAccessorDocComment(printer, descriptor_, CLEARER,
-                               context_->options(),
-                               /* builder */ false, /* kdoc */ true);
-  printer->Print(variables_,
-                 "public fun ${$clear$kt_capitalized_name$$}$() {\n"
-                 "  $kt_dsl_builder$.${$clear$capitalized_name$$}$()\n"
-                 "}\n");
-
-  if (descriptor_->has_presence()) {
-    WriteFieldAccessorDocComment(printer, descriptor_, HAZZER,
-                                 context_->options(),
-                                 /* builder */ false, /* kdoc */ true);
-    printer->Print(
-        variables_,
-        "public fun ${$has$kt_capitalized_name$$}$(): kotlin.Boolean {\n"
-        "  return $kt_dsl_builder$.${$has$capitalized_name$$}$()\n"
-        "}\n");
-  }
-}
-
 void ImmutableEnumFieldGenerator::GenerateFieldBuilderInitializationCode(
     io::Printer* printer) const {
   // noop for enums
@@ -330,19 +277,21 @@ void ImmutableEnumFieldGenerator::GenerateBuilderClearCode(
 void ImmutableEnumFieldGenerator::GenerateMergingCode(
     io::Printer* printer) const {
   if (descriptor_->has_presence()) {
-    printer->Print(variables_,
-                   "if (other.has$capitalized_name$()) {\n"
-                   "  set$capitalized_name$(other.get$capitalized_name$());\n"
-                   "}\n");
-  } else if (SupportUnknownEnumValue(descriptor_)) {
+    printer->Print(variables_, "if (other.has$capitalized_name$()) {\n");
+  } else {
+    printer->Print(variables_, "if (other.$name$_ != $default_number$) {\n");
+  }
+  printer->Indent();
+  if (SupportUnknownEnumValue(descriptor_)) {
     printer->Print(
         variables_,
-        "if (other.$name$_ != $default_number$) {\n"
-        "  set$capitalized_name$Value(other.get$capitalized_name$Value());\n"
-        "}\n");
+        "set$capitalized_name$Value(other.get$capitalized_name$Value());\n");
   } else {
-    ABSL_LOG(FATAL) << "Can't reach here.";
+    printer->Print(variables_,
+                   "set$capitalized_name$(other.get$capitalized_name$());\n");
   }
+  printer->Outdent();
+  printer->Print("}\n");
 }
 
 void ImmutableEnumFieldGenerator::GenerateBuildingCode(
@@ -687,13 +636,13 @@ void RepeatedImmutableEnumFieldGenerator::GenerateMembers(
   printer->Print(
       variables_,
       "@SuppressWarnings(\"serial\")\n"
-      "private java.util.List<java.lang.Integer> $name$_;\n"
+      "private com.google.protobuf.Internal.IntList $name$_;\n"
       "private static final "
-      "com.google.protobuf.Internal.ListAdapter.Converter<\n"
-      "    java.lang.Integer, $type$> $name$_converter_ =\n"
-      "        new com.google.protobuf.Internal.ListAdapter.Converter<\n"
-      "            java.lang.Integer, $type$>() {\n"
-      "          public $type$ convert(java.lang.Integer from) {\n"
+      "com.google.protobuf.Internal.IntListAdapter.IntConverter<\n"
+      "    $type$> $name$_converter_ =\n"
+      "        new com.google.protobuf.Internal.IntListAdapter.IntConverter<\n"
+      "            $type$>() {\n"
+      "          public $type$ convert(int from) {\n"
       "            $type$ result = $type$.forNumber(from);\n"
       "            return result == null ? $unknown$ : result;\n"
       "          }\n"
@@ -701,14 +650,13 @@ void RepeatedImmutableEnumFieldGenerator::GenerateMembers(
   PrintExtraFieldInfo(variables_, printer);
   WriteFieldAccessorDocComment(printer, descriptor_, LIST_GETTER,
                                context_->options());
-  printer->Print(
-      variables_,
-      "@java.lang.Override\n"
-      "$deprecation$public java.util.List<$type$> "
-      "${$get$capitalized_name$List$}$() {\n"
-      "  return new com.google.protobuf.Internal.ListAdapter<\n"
-      "      java.lang.Integer, $type$>($name$_, $name$_converter_);\n"
-      "}\n");
+  printer->Print(variables_,
+                 "@java.lang.Override\n"
+                 "$deprecation$public java.util.List<$type$> "
+                 "${$get$capitalized_name$List$}$() {\n"
+                 "  return new com.google.protobuf.Internal.IntListAdapter<\n"
+                 "      $type$>($name$_, $name$_converter_);\n"
+                 "}\n");
   printer->Annotate("{", "}", descriptor_);
   WriteFieldAccessorDocComment(printer, descriptor_, LIST_COUNT,
                                context_->options());
@@ -725,7 +673,7 @@ void RepeatedImmutableEnumFieldGenerator::GenerateMembers(
       variables_,
       "@java.lang.Override\n"
       "$deprecation$public $type$ ${$get$capitalized_name$$}$(int index) {\n"
-      "  return $name$_converter_.convert($name$_.get(index));\n"
+      "  return $name$_converter_.convert($name$_.getInt(index));\n"
       "}\n");
   printer->Annotate("{", "}", descriptor_);
   if (SupportUnknownEnumValue(descriptor_)) {
@@ -744,7 +692,7 @@ void RepeatedImmutableEnumFieldGenerator::GenerateMembers(
                    "@java.lang.Override\n"
                    "$deprecation$public int "
                    "${$get$capitalized_name$Value$}$(int index) {\n"
-                   "  return $name$_.get(index);\n"
+                   "  return $name$_.getInt(index);\n"
                    "}\n");
     printer->Annotate("{", "}", descriptor_);
   }
@@ -760,19 +708,18 @@ void RepeatedImmutableEnumFieldGenerator::GenerateBuilderMembers(
       variables_,
       // One field is the list and the other field keeps track of whether the
       // list is immutable. If it's immutable, the invariant is that it must
-      // either an instance of Collections.emptyList() or it's an ArrayList
-      // wrapped in a Collections.unmodifiableList() wrapper and nobody else has
-      // a reference to the underlying ArrayList. This invariant allows us to
-      // share instances of lists between protocol buffers avoiding expensive
-      // memory allocations. Note, immutable is a strong guarantee here -- not
-      // just that the list cannot be modified via the reference but that the
-      // list can never be modified.
-      "private java.util.List<java.lang.Integer> $name$_ =\n"
-      "  java.util.Collections.emptyList();\n"
+      // either an instance of emptyIntList() or it's an immutable IntArrayList
+      // and nobody else has a reference to the underlying ArrayList. This
+      // invariant allows us to share instances of lists between protocol
+      // buffers avoiding expensive memory allocations. Note, immutable is a
+      // strong guarantee here -- not just that the list cannot be modified via
+      // the reference but that the list can never be modified.
+      "private com.google.protobuf.Internal.IntList $name$_ =\n"
+      "  emptyIntList();\n"
 
       "private void ensure$capitalized_name$IsMutable() {\n"
       "  if (!$get_mutable_bit_builder$) {\n"
-      "    $name$_ = new java.util.ArrayList<java.lang.Integer>($name$_);\n"
+      "    $name$_ = makeMutableCopy($name$_);\n"
       "    $set_mutable_bit_builder$;\n"
       "  }\n"
       "}\n");
@@ -787,8 +734,8 @@ void RepeatedImmutableEnumFieldGenerator::GenerateBuilderMembers(
       //   immutable.
       "$deprecation$public java.util.List<$type$> "
       "${$get$capitalized_name$List$}$() {\n"
-      "  return new com.google.protobuf.Internal.ListAdapter<\n"
-      "      java.lang.Integer, $type$>($name$_, $name$_converter_);\n"
+      "  return new com.google.protobuf.Internal.IntListAdapter<\n"
+      "      $type$>($name$_, $name$_converter_);\n"
       "}\n");
   printer->Annotate("{", "}", descriptor_);
   WriteFieldAccessorDocComment(printer, descriptor_, LIST_COUNT,
@@ -804,7 +751,7 @@ void RepeatedImmutableEnumFieldGenerator::GenerateBuilderMembers(
   printer->Print(
       variables_,
       "$deprecation$public $type$ ${$get$capitalized_name$$}$(int index) {\n"
-      "  return $name$_converter_.convert($name$_.get(index));\n"
+      "  return $name$_converter_.convert($name$_.getInt(index));\n"
       "}\n");
   printer->Annotate("{", "}", descriptor_);
   WriteFieldAccessorDocComment(printer, descriptor_, LIST_INDEXED_SETTER,
@@ -817,7 +764,7 @@ void RepeatedImmutableEnumFieldGenerator::GenerateBuilderMembers(
                  "    throw new NullPointerException();\n"
                  "  }\n"
                  "  ensure$capitalized_name$IsMutable();\n"
-                 "  $name$_.set(index, value.getNumber());\n"
+                 "  $name$_.setInt(index, value.getNumber());\n"
                  "  onChanged();\n"
                  "  return this;\n"
                  "}\n");
@@ -832,7 +779,7 @@ void RepeatedImmutableEnumFieldGenerator::GenerateBuilderMembers(
                  "    throw new NullPointerException();\n"
                  "  }\n"
                  "  ensure$capitalized_name$IsMutable();\n"
-                 "  $name$_.add(value.getNumber());\n"
+                 "  $name$_.addInt(value.getNumber());\n"
                  "  onChanged();\n"
                  "  return this;\n"
                  "}\n");
@@ -845,7 +792,7 @@ void RepeatedImmutableEnumFieldGenerator::GenerateBuilderMembers(
                  "    java.lang.Iterable<? extends $type$> values) {\n"
                  "  ensure$capitalized_name$IsMutable();\n"
                  "  for ($type$ value : values) {\n"
-                 "    $name$_.add(value.getNumber());\n"
+                 "    $name$_.addInt(value.getNumber());\n"
                  "  }\n"
                  "  onChanged();\n"
                  "  return this;\n"
@@ -857,7 +804,7 @@ void RepeatedImmutableEnumFieldGenerator::GenerateBuilderMembers(
   printer->Print(
       variables_,
       "$deprecation$public Builder ${$clear$capitalized_name$$}$() {\n"
-      "  $name$_ = java.util.Collections.emptyList();\n"
+      "  $name$_ = emptyIntList();\n"
       "  $clear_mutable_bit_builder$;\n"
       "  onChanged();\n"
       "  return this;\n"
@@ -878,7 +825,7 @@ void RepeatedImmutableEnumFieldGenerator::GenerateBuilderMembers(
     printer->Print(variables_,
                    "$deprecation$public int "
                    "${$get$capitalized_name$Value$}$(int index) {\n"
-                   "  return $name$_.get(index);\n"
+                   "  return $name$_.getInt(index);\n"
                    "}\n");
     printer->Annotate("{", "}", descriptor_);
     WriteFieldEnumValueAccessorDocComment(
@@ -889,7 +836,7 @@ void RepeatedImmutableEnumFieldGenerator::GenerateBuilderMembers(
         "$deprecation$public Builder ${$set$capitalized_name$Value$}$(\n"
         "    int index, int value) {\n"
         "  ensure$capitalized_name$IsMutable();\n"
-        "  $name$_.set(index, value);\n"
+        "  $name$_.setInt(index, value);\n"
         "  onChanged();\n"
         "  return this;\n"
         "}\n");
@@ -901,7 +848,7 @@ void RepeatedImmutableEnumFieldGenerator::GenerateBuilderMembers(
                    "$deprecation$public Builder "
                    "${$add$capitalized_name$Value$}$(int value) {\n"
                    "  ensure$capitalized_name$IsMutable();\n"
-                   "  $name$_.add(value);\n"
+                   "  $name$_.addInt(value);\n"
                    "  onChanged();\n"
                    "  return this;\n"
                    "}\n");
@@ -915,7 +862,7 @@ void RepeatedImmutableEnumFieldGenerator::GenerateBuilderMembers(
         "    java.lang.Iterable<java.lang.Integer> values) {\n"
         "  ensure$capitalized_name$IsMutable();\n"
         "  for (int value : values) {\n"
-        "    $name$_.add(value);\n"
+        "    $name$_.addInt(value);\n"
         "  }\n"
         "  onChanged();\n"
         "  return this;\n"
@@ -931,13 +878,13 @@ void RepeatedImmutableEnumFieldGenerator::
 
 void RepeatedImmutableEnumFieldGenerator::GenerateInitializationCode(
     io::Printer* printer) const {
-  printer->Print(variables_, "$name$_ = java.util.Collections.emptyList();\n");
+  printer->Print(variables_, "$name$_ = emptyIntList();\n");
 }
 
 void RepeatedImmutableEnumFieldGenerator::GenerateBuilderClearCode(
     io::Printer* printer) const {
   printer->Print(variables_,
-                 "$name$_ = java.util.Collections.emptyList();\n"
+                 "$name$_ = emptyIntList();\n"
                  "$clear_mutable_bit_builder$;\n");
 }
 
@@ -965,13 +912,12 @@ void RepeatedImmutableEnumFieldGenerator::GenerateBuildingCode(
     io::Printer* printer) const {
   // The code below ensures that the result has an immutable list. If our
   // list is immutable, we can just reuse it. If not, we make it immutable.
-  printer->Print(
-      variables_,
-      "if ($get_mutable_bit_builder$) {\n"
-      "  $name$_ = java.util.Collections.unmodifiableList($name$_);\n"
-      "  $clear_mutable_bit_builder$;\n"
-      "}\n"
-      "result.$name$_ = $name$_;\n");
+  printer->Print(variables_,
+                 "if ($get_mutable_bit_builder$) {\n"
+                 "  $name$_.makeImmutable();\n"
+                 "  $clear_mutable_bit_builder$;\n"
+                 "}\n"
+                 "result.$name$_ = $name$_;\n");
 }
 
 void RepeatedImmutableEnumFieldGenerator::GenerateBuilderParsingCode(
@@ -981,7 +927,7 @@ void RepeatedImmutableEnumFieldGenerator::GenerateBuilderParsingCode(
     printer->Print(variables_,
                    "int tmpRaw = input.readEnum();\n"
                    "ensure$capitalized_name$IsMutable();\n"
-                   "$name$_.add(tmpRaw);\n");
+                   "$name$_.addInt(tmpRaw);\n");
   } else {
     printer->Print(variables_,
                    "int tmpRaw = input.readEnum();\n"
@@ -991,7 +937,7 @@ void RepeatedImmutableEnumFieldGenerator::GenerateBuilderParsingCode(
                    "  mergeUnknownVarintField($number$, tmpRaw);\n"
                    "} else {\n"
                    "  ensure$capitalized_name$IsMutable();\n"
-                   "  $name$_.add(tmpRaw);\n"
+                   "  $name$_.addInt(tmpRaw);\n"
                    "}\n");
   }
 }
@@ -1022,12 +968,12 @@ void RepeatedImmutableEnumFieldGenerator::GenerateSerializationCode(
                    "  output.writeUInt32NoTag($name$MemoizedSerializedSize);\n"
                    "}\n"
                    "for (int i = 0; i < $name$_.size(); i++) {\n"
-                   "  output.writeEnumNoTag($name$_.get(i));\n"
+                   "  output.writeEnumNoTag($name$_.getInt(i));\n"
                    "}\n");
   } else {
     printer->Print(variables_,
                    "for (int i = 0; i < $name$_.size(); i++) {\n"
-                   "  output.writeEnum($number$, $name$_.get(i));\n"
+                   "  output.writeEnum($number$, $name$_.getInt(i));\n"
                    "}\n");
   }
 }
@@ -1042,7 +988,7 @@ void RepeatedImmutableEnumFieldGenerator::GenerateSerializedSizeCode(
   printer->Print(variables_,
                  "for (int i = 0; i < $name$_.size(); i++) {\n"
                  "  dataSize += com.google.protobuf.CodedOutputStream\n"
-                 "    .computeEnumSizeNoTag($name$_.get(i));\n"
+                 "    .computeEnumSizeNoTag($name$_.getInt(i));\n"
                  "}\n");
   printer->Print("size += dataSize;\n");
   if (descriptor_->is_packed()) {
@@ -1080,107 +1026,6 @@ void RepeatedImmutableEnumFieldGenerator::GenerateHashCode(
                  "}\n");
 }
 
-void RepeatedImmutableEnumFieldGenerator::GenerateKotlinDslMembers(
-    io::Printer* printer) const {
-  printer->Print(
-      variables_,
-      "/**\n"
-      " * An uninstantiable, behaviorless type to represent the field in\n"
-      " * generics.\n"
-      " */\n"
-      "@kotlin.OptIn"
-      "(com.google.protobuf.kotlin.OnlyForUseByGeneratedProtoCode::class)\n"
-      "public class ${$$kt_capitalized_name$Proxy$}$ private constructor()"
-      " : com.google.protobuf.kotlin.DslProxy()\n");
-
-  WriteFieldDocComment(printer, descriptor_, context_->options(),
-                       /* kdoc */ true);
-  printer->Print(variables_,
-                 "$kt_deprecation$public val $kt_name$: "
-                 "com.google.protobuf.kotlin.DslList"
-                 "<$kt_type$, ${$$kt_capitalized_name$Proxy$}$>\n"
-                 "  @kotlin.jvm.JvmSynthetic\n"
-                 "  get() = com.google.protobuf.kotlin.DslList(\n"
-                 "    $kt_dsl_builder$.${$get$capitalized_name$List$}$()\n"
-                 "  )\n");
-
-  WriteFieldAccessorDocComment(printer, descriptor_, LIST_ADDER,
-                               context_->options(),
-                               /* builder */ false, /* kdoc */ true);
-  printer->Print(variables_,
-                 "@kotlin.jvm.JvmSynthetic\n"
-                 "@kotlin.jvm.JvmName(\"add$kt_capitalized_name$\")\n"
-                 "public fun com.google.protobuf.kotlin.DslList"
-                 "<$kt_type$, ${$$kt_capitalized_name$Proxy$}$>."
-                 "add(value: $kt_type$) {\n"
-                 "  $kt_dsl_builder$.${$add$capitalized_name$$}$(value)\n"
-                 "}");
-
-  WriteFieldAccessorDocComment(printer, descriptor_, LIST_ADDER,
-                               context_->options(),
-                               /* builder */ false, /* kdoc */ true);
-  printer->Print(variables_,
-                 "@kotlin.jvm.JvmSynthetic\n"
-                 "@kotlin.jvm.JvmName(\"plusAssign$kt_capitalized_name$\")\n"
-                 "@Suppress(\"NOTHING_TO_INLINE\")\n"
-                 "public inline operator fun com.google.protobuf.kotlin.DslList"
-                 "<$kt_type$, ${$$kt_capitalized_name$Proxy$}$>."
-                 "plusAssign(value: $kt_type$) {\n"
-                 "  add(value)\n"
-                 "}");
-
-  WriteFieldAccessorDocComment(printer, descriptor_, LIST_MULTI_ADDER,
-                               context_->options(),
-                               /* builder */ false, /* kdoc */ true);
-  printer->Print(variables_,
-                 "@kotlin.jvm.JvmSynthetic\n"
-                 "@kotlin.jvm.JvmName(\"addAll$kt_capitalized_name$\")\n"
-                 "public fun com.google.protobuf.kotlin.DslList"
-                 "<$kt_type$, ${$$kt_capitalized_name$Proxy$}$>."
-                 "addAll(values: kotlin.collections.Iterable<$kt_type$>) {\n"
-                 "  $kt_dsl_builder$.${$addAll$capitalized_name$$}$(values)\n"
-                 "}");
-
-  WriteFieldAccessorDocComment(printer, descriptor_, LIST_MULTI_ADDER,
-                               context_->options(),
-                               /* builder */ false, /* kdoc */ true);
-  printer->Print(
-      variables_,
-      "@kotlin.jvm.JvmSynthetic\n"
-      "@kotlin.jvm.JvmName(\"plusAssignAll$kt_capitalized_name$\")\n"
-      "@Suppress(\"NOTHING_TO_INLINE\")\n"
-      "public inline operator fun com.google.protobuf.kotlin.DslList"
-      "<$kt_type$, ${$$kt_capitalized_name$Proxy$}$>."
-      "plusAssign(values: kotlin.collections.Iterable<$kt_type$>) {\n"
-      "  addAll(values)\n"
-      "}");
-
-  WriteFieldAccessorDocComment(printer, descriptor_, LIST_INDEXED_SETTER,
-                               context_->options(),
-                               /* builder */ false, /* kdoc */ true);
-  printer->Print(
-      variables_,
-      "@kotlin.jvm.JvmSynthetic\n"
-      "@kotlin.jvm.JvmName(\"set$kt_capitalized_name$\")\n"
-      "public operator fun com.google.protobuf.kotlin.DslList"
-      "<$kt_type$, ${$$kt_capitalized_name$Proxy$}$>."
-      "set(index: kotlin.Int, value: $kt_type$) {\n"
-      "  $kt_dsl_builder$.${$set$capitalized_name$$}$(index, value)\n"
-      "}");
-
-  WriteFieldAccessorDocComment(printer, descriptor_, CLEARER,
-                               context_->options(),
-                               /* builder */ false, /* kdoc */ true);
-  printer->Print(variables_,
-                 "@kotlin.jvm.JvmSynthetic\n"
-                 "@kotlin.jvm.JvmName(\"clear$kt_capitalized_name$\")\n"
-                 "public fun com.google.protobuf.kotlin.DslList"
-                 "<$kt_type$, ${$$kt_capitalized_name$Proxy$}$>."
-                 "clear() {\n"
-                 "  $kt_dsl_builder$.${$clear$capitalized_name$$}$()\n"
-                 "}");
-}
-
 std::string RepeatedImmutableEnumFieldGenerator::GetBoxedType() const {
   return name_resolver_->GetImmutableClassName(descriptor_->enum_type());
 }
